import raytracer;
import common;
import std.math;
import std.stdio;

// Intersection method return values
const int HIT = 1;              // Ray hit primitive
const int MISS = 0;             // Ray missed primitive
const int INPRIM = -1;          // Ray started inside primitive

class Material {
  this() {
    m_Color.Set( 0.2, 0.2, 0.2 );
    m_Refl = 0;
    m_Diff = 0.2;
    m_Spec = 0.8;
    m_RIndex = 1.5;
  }
  void SetColor( Color a_Color ) {
    m_Color = a_Color;
  }
  Color GetColor() {
    return m_Color;
  }
  void SetDiffuse( float a_Diff ) {
    m_Diff = a_Diff;
  }
  void SetSpecular( float a_Spec ) {
    m_Spec = a_Spec;
  }
  void SetReflection( float a_Refl ) {
    m_Refl = a_Refl;
  }
  void SetRefraction( float a_Refr ) {
    m_Refr = a_Refr;
  }
  float GetSpecular() {
    return m_Spec;
  }
  float GetDiffuse() {
    return m_Diff;
  }
  float GetReflection() {
    return m_Refl;
  }
  float GetRefraction() {
    return m_Refr;
  }
  void SetRefrIndex( float a_Refr ) {
    m_RIndex = a_Refr;
  }
  float GetRefrIndex() {
    return m_RIndex;
  }

 private:
  Color m_Color;
  float m_Refl, m_Refr;
  float m_Diff, m_Spec;
  float m_RIndex;
}

class Primitive {
  enum {
    SPHERE = 1,
    PLANE
  }
  this() { m_Light = false; m_Material = new Material; }
  Material GetMaterial() {
    return m_Material;
  }
  void SetMaterial( Material a_Mat ) {
    m_Material = a_Mat;
  }
  abstract int GetType();
  abstract int Intersect( inout Ray a_Ray, inout float a_Dist );
  abstract vector3 GetNormal( vector3 a_Pos );
  Color GetColor( vector3 ) {
    return m_Material.GetColor();
  }
  void Light( bool a_Light ) {
    m_Light = a_Light;
  }
  bool IsLight() {
    return m_Light;
  }
  void SetName( char[] a_Name ) {
    m_Name = a_Name.dup;
  }
  char[] GetName() {
    return m_Name.dup;
  }

 protected:
  Material m_Material;
  char[] m_Name;
  bool m_Light;
}

class Sphere : Primitive {

  int GetType() {
    return SPHERE;
  }
  this( vector3 a_Centre, float a_Radius ) {
    m_Centre = a_Centre;
    m_SqRadius = a_Radius * a_Radius;
    m_Radius = a_Radius;
    m_RRadius = 1 / a_Radius;
  }
  vector3 GetCentre() {
    return m_Centre;
  }
  float GetSqRadius() {
    return m_SqRadius;
  }
  vector3 GetNormal( vector3 a_Pos ) {
    return (a_Pos - m_Centre) * m_RRadius;
  }

  int Intersect( inout Ray a_Ray, inout float a_Dist ) {
//    vector3 v = a_Ray.origin - m_Centre;
//    float b = -DOT( v, a_Ray.direction );
//    float det = (b * b) - DOT( v, v ) + m_SqRadius;
//    int retval = MISS;
    vector3 v = a_Ray.origin;
    v -= m_Centre;
    vector3 dir = a_Ray.direction;
    float b = -(v.x * dir.x + v.y * dir.y + v.z * dir.z);
    float det = (b * b) - (v.x*v.x + v.y*v.y + v.z*v.z) + m_SqRadius;
    int retval = MISS;
    
    if (det > 0) {
      det = sqrt( det );
      float i1 = b - det;
      float i2 = b + det;
      if (i2 > 0) {
        if (i1 < 0) {
          if (i2 < a_Dist) {
            a_Dist = i2;
            retval = INPRIM;
          }
        } else {
          if (i1 < a_Dist) {
            a_Dist = i1;
            retval = HIT;
          }
        }
      }
    }
    return retval;
  }


 private:
  vector3 m_Centre;
  float m_SqRadius, m_Radius, m_RRadius;
}

class PlanePrim : public Primitive {

  int GetType() {
    return PLANE;
  }
  this( vector3 a_Normal, float a_D )  { m_Plane = new plane( a_Normal, a_D ); }
  vector3 GetNormal() {
    return m_Plane.N;
  }
  float GetD() {
    return m_Plane.D;
  }

  int Intersect( inout Ray a_Ray, inout float a_Dist ) {
//    float d = DOT( m_Plane.N, a_Ray.direction );
    float d = (m_Plane.N.x*a_Ray.direction.x + m_Plane.N.y*a_Ray.direction.y + m_Plane.N.z*a_Ray.direction.z);

    if (d != 0) {
//      float dist = -(DOT( m_Plane.N, a_Ray.origin ) + m_Plane.D) / d;
      float dist = -((m_Plane.N.x*a_Ray.origin.x + m_Plane.N.y*a_Ray.origin.y + m_Plane.N.z*a_Ray.origin.z) + m_Plane.D) / d;
      if (dist > 0) {
        if (dist < a_Dist) {
          a_Dist = dist;
          return HIT;
        }
      }
    }
    return MISS;
  }

  vector3 GetNormal( vector3 a_Pos ) {
    return m_Plane.N;
  }


 private:
  plane m_Plane;
}

class Scene {
  this() { }
  int opApply(int delegate(inout Primitive) dg) {
    int result = 0;
    foreach (Primitive p; m_Primitives) {
      result = dg(p);
      if (result) {
        break;
      }
    }
    return result;
  }
  void SphereTree( inout int a_Prim, float a_Radius, vector3 a_Pos, int a_Depth ) {
  
  }
  void InitScene() {
    Primitive prim;

    // ground plane
    prim = new PlanePrim( vector3( 0, 1, 0 ), 4.4f );
    prim.SetName( "plane" );
    prim.GetMaterial().SetReflection( 0 );
    prim.GetMaterial().SetRefraction( 0 );
    prim.GetMaterial().SetDiffuse( 1.0f );
    prim.GetMaterial().SetColor( Color( 0.4f, 0.3f, 0.3f ) );
    m_Primitives ~= prim;
    // big sphere
    prim = new Sphere( vector3( 2, 0.8f, 3 ), 2.5f );
    prim.SetName( "big sphere" );
    prim.GetMaterial().SetReflection( 0.2f );
    prim.GetMaterial().SetRefraction( 0.8f );
    prim.GetMaterial().SetRefrIndex( 1.3f );
    prim.GetMaterial().SetColor( Color( 0.7f, 0.7f, 1.0f ) );
    m_Primitives ~= prim;
    // small sphere
    prim = new Sphere( vector3( -5.5f, -0.5, 7 ), 2 );
    prim.SetName( "small sphere" );
    prim.GetMaterial().SetReflection( 0.5f );
    prim.GetMaterial().SetRefraction( 0.0f );
    prim.GetMaterial().SetRefrIndex( 1.3f );
    prim.GetMaterial().SetDiffuse( 0.1f );
    prim.GetMaterial().SetColor( Color( 0.7f, 0.7f, 1.0f ) );
    m_Primitives ~= prim;
    // light source 1
    prim = new Sphere( vector3( 0, 5, 5 ), 0.1f );
    prim.Light( true );
    prim.GetMaterial().SetColor( Color( 0.4f, 0.4f, 0.4f ) );
    m_Primitives ~= prim;
    // light source 2
    prim = new Sphere( vector3( -3, 5, 1 ), 0.1f );
    prim.Light( true );
    prim.GetMaterial().SetColor( Color( 0.6f, 0.6f, 0.8f ) );
    m_Primitives ~= prim;
    // extra sphere
    prim = new Sphere( vector3( -1.5f, -3.8f, 1 ), 1.5f );
    prim.SetName( "extra sphere" );
    prim.GetMaterial().SetReflection( 0.0f );
    prim.GetMaterial().SetRefraction( 0.8f );
    prim.GetMaterial().SetColor( Color( 1.0f, 0.4f, 0.4f ) );
    m_Primitives ~= prim;
    // back plane
    prim = new PlanePrim( vector3( 0.4f, 0, -1 ), 12 );
    prim.SetName( "back plane" );
    prim.GetMaterial().SetReflection( 0.0f );
    prim.GetMaterial().SetRefraction( 0.0f );
    prim.GetMaterial().SetSpecular( 0 );
    prim.GetMaterial().SetDiffuse( 0.6f );
    prim.GetMaterial().SetColor( Color( 0.5f, 0.3f, 0.5f ) );
    m_Primitives ~= prim;
    // ceiling plane
    prim = new PlanePrim( vector3( 0, -1, 0 ), 7.4f );
    prim.SetName( "back plane" );
    prim.GetMaterial().SetReflection( 0.0f );
    prim.GetMaterial().SetRefraction( 0.0f );
    prim.GetMaterial().SetSpecular( 0 );
    prim.GetMaterial().SetDiffuse( 0.5f );
    prim.GetMaterial().SetColor( Color( 0.4f, 0.7f, 0.7f ) );
    m_Primitives ~= prim;
    // grid
    for ( int x = 0; x < 8; x++ ) {
      for ( int y = 0; y < 7; y++ ) {
        prim = new Sphere( vector3( -4.5f + x * 1.5f, -4.3f + y * 1.5f, 10 ), 0.3f );
        prim.SetName( "grid sphere" );
        prim.GetMaterial().SetReflection( 0 );
        prim.GetMaterial().SetRefraction( 0 );
        prim.GetMaterial().SetSpecular( 0.6f );
        prim.GetMaterial().SetDiffuse( 0.6f );
        prim.GetMaterial().SetColor( Color( 0.3f, 1.0f, 0.4f ) );
        m_Primitives ~= prim;
      }
    }
  }

// private:
  Primitive[] m_Primitives;
}


